package doctrans;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileInputStream;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Hashtable;
import org.sanskritlibrary.IOAssistance;
import org.sanskritlibrary.Regex;
import org.sanskritlibrary.FSM;
import org.sanskritlibrary.ResourceInputStream;
import org.sanskritlibrary.webservice.TransformMap;
import org.sanskritlibrary.webservice.WebServices;

// Converts RTF to 
public class Main implements ResourceInputStream
{
    static Hashtable<String, String> outerMap = new Hashtable<String, String>();
    static Hashtable<String, String> innerMap = new Hashtable<String, String>();
    static String defaultForeign = "<foreign xml:lang='sa-Latn-x-SLP1'>%s</foreign>";
    static Hashtable<String, FSM[]> fsmh = null;

    Hashtable<String, String> newcommands = new Hashtable<String, String>();
    TransformMap tmTexRomanSLP1 = new TransformMap("*:texRoman->SLP1", this, fsmh);
    TransformMap tmSLP1RomanUnicode = new TransformMap("*:SLP1->roman", this, fsmh);
    static
    {
        outerMap.put("\\textit", "<foreign xml:lang='sa-Latn-x-SLP1' rend='italics'>%s</foreign>");
        outerMap.put("\\textbf", "<seg rend='bold'>%s</seg>");

        innerMap.put("\\textit", "<seg rend='italics'>%s</seg>");
        innerMap.put("\\textbf", "<seg rend='bold'>%s</seg>");
    }

    int chno = 0;

    public static void main(String[] args) throws Exception
    {
        new Main(args);
    }


    public Main(String[] args) throws Exception
    {
        // Pass 1 read newcommands
        pass1(args[0], true);

        // Pass 2 replace all macros
        pass2(args[0]);

        // Pass 3 do transcodings and add in <i>, <b>
        pass3();

        // Pass 4 convert <i>roman unicode</> to RTF
        pass4("out.xml");

        pass5("out.rtf");
    }


    private void pass1(String filename, boolean fixup) throws Exception
    {
        BufferedWriter bw = IOAssistance.openWriter("temp.slp1");
        BufferedReader br = IOAssistance.openReader(filename);
        String line = null;
        Regex re = new Regex("", "^\\\\newcommand\\s*\\{(.+?)\\}\\s*\\{(.+)\\}(.*)");

        // find all newcommands and put them in hash table
        while ((line = br.readLine())!=null)
        {
            re.reset(line);
            if (re.find())
            {
                String command = re.getGroup(1);
                String text = re.getGroup(2);
                String comment = re.getGroup(3);
                if (fixup)
                    text = transcode(text, "<s>", "</s>");
                bw.write(String.format("\\newcommand{%s}{%s}%s\n", command, text, comment));
                newcommands.put(command, text);
            }
        }

        bw.close();
        br.close();
    }

    private void pass2(String filename) throws Exception
    {
        char ch = 0;
        //BufferedReader br = IOAssistance.openReader(filename);
        String text = IOAssistance.readEntireFile(filename);

        StringBuilder command = new StringBuilder();
        boolean inCommand = false;
        boolean ignore = false;
        StringBuilder buffer = new StringBuilder();
        StringBuilder foreign = new StringBuilder();
        boolean inTag = false;
        int len = text.length();
        for (int i=0; i<len; i++)
        {
            ch = text.charAt(i);

            int j = tagged(text, i, inTag);
            if (j != 0)
            {
                if (inTag)
                {
                    buffer.append(WebServices.transformString(foreign.toString(), tmTexRomanSLP1));
                    foreign.setLength(0);
                    buffer.append(text.substring(i, j));
                    inTag = false;
                }
                else
                {
                    buffer.append(text.substring(i, j));
                    inTag = true;
                }

                i = j-1;
            }
            else if (inTag)
                foreign.append(ch);
            else if (ignore)           // send newcommands through
            {
                if (ch == '\n')
                    ignore = false;
            }
            else if (ch == '\\')
            {
                if (inCommand && command.length() == 1 && command.charAt(0) == '\\')
                    command.append((char)ch);             // handle \\ which Peter uses for end of line marker
                else
                {
                    command.setLength(0);
                    command.append((char)ch);
                    inCommand = true;
                }
            }
            else if (inCommand)
            {
                if (Character.isLetterOrDigit(ch))
                    command.append((char)ch);
                else
                {
                    inCommand = false;
                    String key = command.toString();
                    String value = newcommands.get(key);
                    if (value != null)          // replace macro
                    {
                       buffer.append(value);
                       buffer.append((char)ch);
                    }
                    else
                    {
                        if (key.equals("\\newcommand"))
                            ignore = true;
                        else
                        {
                            buffer.append(key);
                            buffer.append((char)ch);
                        }
                    }
                }
            }
            else
                buffer.append((char)ch);
        }

        BufferedWriter bw = IOAssistance.openWriter("out.tmp");
        String _buffer = buffer.toString();
        len = _buffer.length();
        char prevch = 0;
        int braceLevel = 0;
        // add <s>...</s> tags to all \textit or \textbf (note: there are no \newcommands to deal with
        for (int i=0; i<len; i++)
        {
            ch = _buffer.charAt(i);
            if (ch == '\\')
            {
                String s = _buffer.substring(i+1);
                if (ignore)
                {
                    bw.write(ch);
                    if (ch == '\n')
                        ignore = false;
                }
                else if (s.startsWith("newcommand"))
                {
                    ignore = true;
                    bw.write(ch);
                }
                else if (s.startsWith("textit") || s.startsWith("textbf"))
                {
                    bw.write("\\");
                    bw.write(s.substring(0, 6));
                    i+=7;           // move past \textit or \textbf
                    int j = findEndBrace(_buffer, i);
                    s = _buffer.substring(i+1, j);
                    bw.write("{");
                    if (!s.startsWith("<s>"))
                    {
                        /*
                        bw.write("<s>");
                        bw.write(WebServices.transformString(s, tmTexRomanSLP1));
                        bw.write("</s>");
                         */
                        bw.write(transcode(s, "<s>", "</s>"));
                    }
                    else
                        bw.write(s);
                    bw.write("}");
                    i = j;              // this will be the }
                }
            }
            else
                bw.write(ch);
        }
        bw.flush();
        bw.close();

        // Now convert {<s>...</s>} to <s>...</s>
        _buffer = IOAssistance.readEntireFile("out.tmp");
        bw = IOAssistance.openWriter("out.tmp");
        len = _buffer.length();
        prevch = 0;
        braceLevel = 0;
        for (int i=0; i<len; i++)
        {
            ch = _buffer.charAt(i);
            if (ch == '{')
            {
                if (prevch <= ' ' || prevch == '-')
                    braceLevel = 1;
                else if (braceLevel > 0)
                {
                    braceLevel++;
                    bw.write(ch);
                }
                else
                    bw.write(ch);
            }
            else if (ch == '}')
            {
                if (braceLevel == 0)
                    bw.write(ch);
                else
                {
                    // do not write out '}' at level 1
                    if (braceLevel > 1)
                        bw.write(ch);
                    braceLevel--;
                }
            }
            else
                bw.write(ch);

            prevch = (char)ch;
        }
        //bw.write(buffer.toString());
        bw.flush();
        bw.close();
    }


    private void pass3() throws Exception
    {
        String endtag = null;
        ArrayList<String> endtags = new ArrayList<String>();
        StringBuilder buffer = new StringBuilder();
        String text = IOAssistance.readEntireFile("out.tmp");
        BufferedWriter bw = IOAssistance.openWriter("out.xml");
        write(bw, "<tex>\n");
        boolean inTag = false;
        int len = text.length();
        for (int i=0; i<len; i++)
        {
            if (i == 1010)
                i = 1010;
            char ch = text.charAt(i);
            if (ch == '!')
                ch = '!';
            int j = tagged(text, i, inTag);
            if (j != 0)
            {
                if (inTag)
                {
                    bw.write(WebServices.transformString(buffer.toString(), tmSLP1RomanUnicode));
                    buffer.setLength(0);
                    inTag = false;
                }
                else
                    inTag = true;

                i = j-1;
            }
            else if (inTag)
                buffer.append(ch);
            else if (ch == '\\')            // tex tag
            {
                if (text.charAt(i+1) == '\\')       // it's an end of line marker
                {
                    bw.write("\\\\");
                    i++;
                    continue;
                }

                j = getCommand(text, i);
                String command = text.substring(i, j);
                if (command.equals("\\textit"))
                {
                    bw.write("<i>");
                    endtags.add("</i>");
                    j++;        // skip the { so it is not written
                }
                else if (command.equals("\\textbf"))
                {
                    bw.write("<b>");
                    endtags.add("</b>");
                    j++;        // skip the { so it is not written
                }
                else
                {
                    bw.write(command);
                    if (text.charAt(j) == '{')
                        endtag = "}";
                }

                i = j-1;
            }
            else if (ch == '{')
            {
                if (endtag == null)
                    endtag = "}";
                endtags.add(endtag);
                endtag = null;
                bw.write(ch);
            }
            else if (ch == '}')
            {
                int last = endtags.size()-1;
                if (last == -1)
                    last = -1;
                else
                {
                    bw.write(endtags.get(last));
                    endtags.remove(last);
                }
            }
            else
                write(bw, ch);
        }
        write(bw, "</tex>\n");
        bw.flush();
        bw.close();

        String all = IOAssistance.readEntireFile("out.xml");
        all = all.replaceAll("\\{<b>", "<b>");
        all = all.replaceAll("</b>\\}", "</b>");
        all = all.replaceAll("\\{<i>", "<i>");
        all = all.replaceAll("</i>\\}", "</i>");
        bw = IOAssistance.openWriter("out.xml");
        bw.write(all);
        bw.close();
    }

    int tagged(String text, int i, boolean inTag)
    {
        String s = text.substring(i);
        if (!inTag)
        {
            if (s.startsWith("<s>"))
                return i+3;
        }
        else if (s.startsWith("</s>"))
            return i+4;

        return 0;
    }

    int getCommand(String text, int i)
    {
        int j = 0;
        char ch = 0;
        int len = text.length();
        for (j=i+1; j<len; j++)
        {
            ch = text.charAt(j);
            if (!Character.isLetter(ch))
                break;
        }

        return j;
    }

    private void pass4(String filename) throws Exception
    {
        BufferedWriter bw = IOAssistance.openWriter("out.rtf");

        new ParseHTML(filename, bw, tmSLP1RomanUnicode);

        bw.close();
    }


    private void pass5(String filename) throws Exception
    {
        BufferedWriter bw = IOAssistance.openWriter("outrtf.txt");

        new ParseUnicodeRoman(filename, bw);

        bw.close();
    }


    private void write(BufferedWriter bw, char ch) throws Exception
    {
        if (ch == '&')
            bw.write("&amp;");
        else
            bw.write(ch);
    }

    private void write(BufferedWriter bw, String s) throws Exception
    {
        bw.write(s.replaceAll("&", "&amp;"));
    }


    public String transcode(String text, String starttag, String endtag) throws Exception
    {
        StringBuilder sb = new StringBuilder();
        StringBuilder buffer = new StringBuilder();
        boolean[] bufferWantsBrace = new boolean[5];
        int braceLevel = 0;
        int len = text.length();

        for (int i=0; i<len; i++)
        {
            char ch = text.charAt(i);
            if (ch == '\\')
            {
                int j = texCommand(text, i);
                if (j != 0)
                {
                    if (buffer.length() > 0)
                    {
                        sb.append(starttag);
                        sb.append(WebServices.transformString(buffer.toString(), tmTexRomanSLP1));
                        sb.append(endtag);
                        buffer.setLength(0);
                    }
                    sb.append(text, i, j);
                    i = j-1;
                    bufferWantsBrace[braceLevel] = false;
                }
                else
                {
                    buffer.append(ch);
                    bufferWantsBrace[braceLevel] = true;
                }
            }
            else if (ch == '{')
            {
                if (bufferWantsBrace[braceLevel])
                    buffer.append(ch);
                else
                    sb.append(ch);
                braceLevel++;
            }
            else if (ch == '}')
            {
                braceLevel --;
                if (bufferWantsBrace[braceLevel])
                    buffer.append(ch);
                else
                {
                    if (buffer.length() > 0)
                    {
                        sb.append("<s>");
                        sb.append(WebServices.transformString(buffer.toString(), tmTexRomanSLP1));
                        sb.append("</s>");
                        buffer.setLength(0);
                    }
                    sb.append(ch);
                }
            }
            else
                buffer.append(ch);
        }

        if (buffer.length() > 0)
        {
            sb.append("<s>");
            sb.append(WebServices.transformString(buffer.toString(), tmTexRomanSLP1));
            sb.append("</s>");
        }


        return sb.toString();
    }

    public int texCommand(String text, int i)
    {
        String s = text.substring(i);
        if (s.startsWith("\\textit"))
            return i+7;
        else if (s.startsWith("\\textbf"))
            return i+7;
        else
            return 0;
    }

    public String transcodeX(String text, int level) throws Exception
    {
        text = fixup(text);
        String template = null;
        boolean inCommand = false;
        StringBuilder cmd = new StringBuilder();
        StringBuilder foreign = new StringBuilder();
        StringBuilder sb = new StringBuilder();
        int len = text.length();
        boolean expectingBrace = false;
        for (int i=0; i<len; i++)
        {
            char ch = text.charAt(i);
            if (ch == '\\')     // either \textit, \textbf or a representation of a text character with diacritics
            {
                inCommand = true;
                cmd.setLength(0);
                cmd.append((char)ch);
            }
            else if (inCommand)
            {
                if (Character.isLetterOrDigit(ch))
                    cmd.append((char)ch);
                else        // at end of command name
                {
                    if (cmd.toString().equals("\\textbf"))
                        inCommand = false;
                    inCommand = false;
                    Hashtable<String, String> map = null;
                    if (level == 0)
                        map = outerMap;
                    else
                        map = innerMap;
                    template = map.get(cmd.toString());
                    if (template == null)
                    {
                        foreign.append(cmd.toString());             // assume that it is command for roman character
                        foreign.append((char)ch);
                        expectingBrace = false;
                    }
                    else        // recursively do the command
                    {
                        if (foreign.length() > 0)
                        {
                            sb.append(WebServices.transformString(foreign.toString(), tmTexRomanSLP1));
                            foreign.setLength(0);
                        }
                        if (ch != '{')
                            System.out.format("Command not followed by {: %s", text);
                        else
                        {
                            int endBrace = findEndBrace(text, i);
                            String innerText = transcodeX(text.substring(i+1, endBrace), level+1);
                            sb.append(String.format(template, innerText));
                            i = endBrace;
                        }
                        expectingBrace = ch != '{';
                    }
                }
            }
            else if (expectingBrace)
            {
                if (ch == '{')
                {
                    foreign.append((char)ch);
                    expectingBrace = false;
                }
                else
                    return "????";
            }
            else
                foreign.append((char)ch);
        }

        if (foreign.length() > 0)
            sb.append(WebServices.transformString(foreign.toString(), tmTexRomanSLP1));

        return sb.toString();
    }

    private String fixup(String text)
    {
        while (text.startsWith("{"))
        {
            int braceLevel = 0;
            int len = text.length();
            for (int i=0; i<len; i++)
            {
                char ch = text.charAt(i);
                if (ch == '{')
                    braceLevel++;
                else if (ch == '}')
                {
                    braceLevel--;
                    if (braceLevel == 0 && i != len-1)         // e.g. have something like {\'s}a\textbf{p} where first { does not belong to last }
                        return text;
                }
            }
            text = text.substring(1, len-1);
        }

        return text;
    }

    private int findEndBrace(String text, int i)
    {
        int braceLevel = 0;

        for (; i<text.length(); i++)
        {
            char ch = text.charAt(i);
            if (ch == '{')
                braceLevel++;
            else if (ch == '}')
            {
                braceLevel--;
                if (braceLevel == 0)
                    return i;
            }
        }

        System.out.format("Missing } in %s\n", text);

        return i;       // should not reach here!
    }

    public InputStream[] getInputStreams(String sourceEncoding, String targetEncoding)
    {
        String[] fileNames = FSM.createFileNames("", sourceEncoding, targetEncoding);
        InputStream[] iss = new InputStream[fileNames.length];
        int i = 0;

        for (String name : fileNames)
        {
            String fname = String.format(".%s", name);
            try
            {
                iss[i++] = new FileInputStream(fname);
            } catch (Exception ex)
            {
            }
        }

        return iss;
    }

}
