package doctrans;

import java.io.BufferedWriter;
import java.io.File;
import java.io.IOException;
import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.helpers.DefaultHandler;

import javax.xml.parsers.SAXParserFactory;
import javax.xml.parsers.SAXParser;
import org.sanskritlibrary.TextFormatter;

import org.sanskritlibrary.webservice.TransformMap;
import org.sanskritlibrary.webservice.WebServices;

public class ParseHTML extends DefaultHandler //implements ErrorHandler, ContentHandler, DTDHandler, EntityResolver
{
    Locator locator;
    BufferedWriter bw;
    String filename;
    StringBuilder buffer = new StringBuilder();
    TransformMap tm;
    int level = 0;
    int textLevel = 0;

    public ParseHTML(String fn, BufferedWriter bw, TransformMap tm)
    {
        this.bw = bw;
        this.tm = tm;
        // Use the default (non-validating) parser
        SAXParserFactory factory = SAXParserFactory.newInstance();
        try
        {
            // Parse the input
            SAXParser saxParser = factory.newSAXParser();
            File f = new File(fn);
            filename = f.getName();
            filename = filename.substring(0, filename.length()-5);
            saxParser.parse(f, this);

        }
        catch (Throwable t)
        {
            t.printStackTrace ();
        }
    }


    // For ErrorHandler
    @Override
    public void warning(SAXParseException exception) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void error(SAXParseException exception) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void fatalError(SAXParseException exception) throws SAXException
    {
        System.out.format(String.format("Error parsing %s.htm: %s\n", filename, exception.getMessage()));
        throw exception;
    }

    // For ContentHandler
    @Override
    public void setDocumentLocator(Locator locator)
    {
        this.locator = locator;
    }

    @Override

    public void startDocument() throws SAXException
    {
        try
        {
            bw.write("{\\rtf1\\mac\\ansicpg10000{\\fonttbl\\f0\\froman\\fcharset77 Times-Roman;\\f1\\fnil\\fcharset77 ManjushreeCSX;\\f2\\fnil\\fcharset77 Vedic;\n}\n{\\colortbl;\\red255\\green255\\blue255;}\n\\margl1440\\margr1440\\vieww22080\\viewh16120\\viewkind0\n\\pard\\ql\\qnatural\\pardirnatural");
        }
        catch(Exception ex)
        {
            throw new SAXException(ex.getMessage());
        }
    }

    @Override
    public void endDocument() throws SAXException
    {
        try
        {
            bw.write("}");
        }
        catch(Exception ex)
        {
            throw new SAXException(ex.getMessage());
        }
    }

    @Override
    public void startPrefixMapping(String prefix, String uri) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void endPrefixMapping(String prefix) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException
    {
        try
        {
            String code = null;
            if (qName.equals("i"))
            {
                code = "{\\i ";
                textLevel++;
            }
            else if (qName.equals("b"))
            {
                code = "{\\b ";
                textLevel++;
            }

            if (code != null)
                bw.write(code);
        }
        catch (Exception ex)
        {
            throw new SAXException(ex.getMessage());
        }
    }


    @Override
    public void endElement(String uri, String localName, String qName) throws SAXException
    {
        try
        {
            if (qName.equals("i") || qName.equals("b"))
            {
                textLevel--;
                bw.write("}");
            }
        }
        catch (Exception ex)
        {
            throw new SAXException(ex.getMessage());
        }
    }

    @Override
    public void characters(char[] ch, int start, int length) throws SAXException
    {
        try
        {
            String s = new String(ch, start, length);
            //s = s.replaceAll("&", "&amp;");
            s = s.replaceAll("\\\\-", "!");
            s = s.replaceAll("\\\\", "\\\\\\\\");
            s = s.replaceAll("!", "\\\\-");
            s = s.replaceAll("\\{", "\\\\{");
            s = s.replaceAll("\\}", "\\\\}");
            s = s.replaceAll("\n", "\\\\par ");

            int len = s.length();
            for (int i=0; i<len; i++)
            {
                char chr = s.charAt(i);
                if (chr <= 0x7f)
                    bw.write(chr);
                else if (chr == 0xAD)
                    bw.write("\\-");
                else
                {
                    bw.write(TextFormatter.charToSlashUDecimal(chr));
                    bw.write("\\'3f");
                }
            }
        }
        catch (Exception ex)
        {
            throw new SAXException(ex.getMessage());
        }
    }

    @Override
    public void ignorableWhitespace(char[] ch, int start, int length) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void processingInstruction(String target, String data) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void skippedEntity(String name) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    // For DTDHandler
    @Override
    public void notationDecl(String name, String publicId, String systemId) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void unparsedEntityDecl(String name, String publicId, String systemId, String notationName) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    // For EntityHandler
    @Override
    public InputSource resolveEntity(String publicId, String systemId) throws SAXException, IOException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }
}
